/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/

package ca.mcgill.cs.stg.fraco;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IMarker;

import ca.mcgill.cs.stg.fraco.Properties.ResultsFiltering;
import ca.mcgill.cs.stg.fraco.rules.Rules;
import ca.mcgill.cs.stg.fraco.rules.RulesFactory;
import ca.mcgill.cs.stg.fraco.rules.ScopeRules;
import ca.mcgill.cs.stg.fraco.utils.Comments;
import ca.mcgill.cs.stg.fraco.utils.FragileComments;
import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.RuleTypes;

/**
 * A singleton class used to control the plugin properties and common place to
 * sequentially call the different classes needed for the processing.
 * 
 * @author inder
 *
 */
public class DetectFragileComments {
	/**
	 * Contains the existing marker ids. Needed to check if the marker already
	 * exists or not while publishing results.
	 */
	public static List<IMarker> _markerIds = new ArrayList<IMarker>();

	public static boolean _identicalMatch;
	public static boolean _lemmaMatch;

	/**
	 * flag used to apply proximity filter on comments.
	 */
	public static boolean _filterProximity = true;

	private static DetectFragileComments _detectFragilityComments = new DetectFragileComments();

	private DetectFragileComments() {
	}

	/**
	 * Starts the process of fragility detection. Applies scope
	 * ca.mcgill.cs.stg.fraco.rules on identifier, processes comments, finds
	 * fragile matches and publishes results.
	 * 
	 * @param id
	 *            renamed identifier
	 * @param commentsTable
	 *            table of comments present in the while
	 */
	static void InitFragilityDetection() {
		// BasicConfigurator.configure();
		ApplyIdScopingRules();
		ApplyCommentScopingRules();
		ApplyMatchingRules();
		PublishResults();
	}

	/**
	 * Identifier is passed to the scopingRules class to find its scope
	 */
	private static void ApplyIdScopingRules() {
		ScopeRules.ApplyScopingRules(AstBuilder._identifier);
	}

	/**
	 * After identifying an identifier's scope, we process every comment in the
	 * source code to get the comments in scope of the identifier.
	 */
	private static void ApplyCommentScopingRules() {
		ResultsFiltering filter = new ResultsFiltering();
		filter.ApplyCommentScopeRules(AstBuilder._commentsTable, AstBuilder._identifier);
	}

	/**
	 * Depending upon the choice of developer, matching
	 * ca.mcgill.cs.stg.fraco.rules are applied. If no property is chose, the
	 * default is lemmas and tokens match rule which incorporates identical and
	 * fuzzy matches.
	 */
	private static void ApplyMatchingRules() {
		// fuzzyMatchRules();
		if (_identicalMatch)
			identicalMatchRules();
		else
			lemmaAndTokensMatchRules();
	}

	/**
	 * This method calls the identical match rule that only finds the identical
	 * matches for an identifier.
	 */
	public synchronized static void identicalMatchRules() {
		List<Comments> matchedComments = new ArrayList<Comments>();

		RulesFactory ruleFactory = new RulesFactory();
		Rules identicalMatch = ruleFactory.getRule(RuleTypes.IDENTICAL_MATCH);
		for (Comments comment : AstBuilder._identifier.get_comments()) {
			if (identicalMatch.ApplyRules(AstBuilder._identifier, comment))
				matchedComments.add(comment);
		}
		AstBuilder._identifier.set_comments(matchedComments);
	}

	/**
	 * This method calls the identical match rule that only finds the fuzzy
	 * matches for an identifier.
	 */
	public synchronized static void fuzzyMatchRules() {
		List<Comments> matchedComments = new ArrayList<Comments>();

		RulesFactory ruleFactory = new RulesFactory();
		Rules fuzzyMatch = ruleFactory.getRule(RuleTypes.FUZZY_MATCH);
		for (Comments comment : AstBuilder._identifier.get_comments()) {

			if (fuzzyMatch.ApplyRules(AstBuilder._identifier, comment))
				matchedComments.add(comment);
		}
		AstBuilder._identifier.set_comments(matchedComments);
	}

	/**
	 * This method calls the identical match rule that finds the identical,
	 * fuzzy, tokens and lemma matches for an identifier.
	 */
	public synchronized static void lemmaAndTokensMatchRules() {
		List<Comments> matchedComments = new ArrayList<Comments>();

		RulesFactory ruleFactory = new RulesFactory();
		Rules lemmaAndTokensMatch = ruleFactory.getRule(RuleTypes.LEMMA_TOKENS_SEMANTIC);
		for (Comments comment : AstBuilder._identifier.get_comments()) {

			if (lemmaAndTokensMatch.ApplyRules(AstBuilder._identifier, comment))
				matchedComments.add(comment);
		}

		AstBuilder._identifier.set_comments(matchedComments);
	}

	/**
	 * This method calls the publishResults method of class ResultsFiltering to
	 * finally create marker for each fragile comment.
	 */
	public static void PublishResults() {

		if (!AstBuilder._identifier.get_comments().isEmpty()) {
			ResultsFiltering results = new ResultsFiltering();
			results.PublishResults(_markerIds, AstBuilder._identifier);
		}
	}
}
