/*******************************************************************************
 * Fraco - Eclipse plug-in to detect fragile comments
 *
 * Copyright (C) 2019 McGill University
 *     
 * Eclipse Public License - v 2.0
 *
 *  THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF THIS ECLIPSE
 *  PUBLIC LICENSE v2.0. ANY USE, REPRODUCTION OR DISTRIBUTION
 *  OF THE PROGRAM CONSTITUTES RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.
 *******************************************************************************/
package ca.mcgill.cs.stg.fraco.parsers;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.core.BinaryMethod;
import org.eclipse.jdt.internal.corext.dom.Bindings;

import ca.mcgill.cs.stg.fraco.AstBuilder;
import ca.mcgill.cs.stg.fraco.utils.Identifier;
import ca.mcgill.cs.stg.fraco.utils.Scope;
import ca.mcgill.cs.stg.fraco.utils.Type;
import ca.mcgill.cs.stg.fraco.visitors.MethodVisitor;

public class MethodParser extends ParseAstNodes {

	/**
	 * Stores the name of the method declaration parsed.
	 */
	private String methodName;

	/**
	 * the name of the formal parameter renamed. It is null if the AST is parsed
	 * only for method declarations, else will have the value of the parameter
	 * which is renamed.
	 */
	private Optional<String> parameterName = Optional.ofNullable(null);

	/**
	 * Class Name of the parent of the method under parsing.
	 */
	private String parentName;

	/**
	 * the method declaration under parsing
	 */
	private MethodDeclaration method;
	/**
	 * Index of the leading method comment in comment list used to store all the
	 * comments in a compilation unit.
	 */
	private int leadingCommentIndex;

	private boolean methodfound;

	/**
	 * Parse all method type AST nodes in compilation Unit.
	 * 
	 * @param parameterName,
	 *            passed name when called from ParseLocalVariables method to get
	 *            leading comment associated with local variable
	 */

	@SuppressWarnings("restriction")
	public void parseMethods(Optional<String> parameterName) {
		MethodVisitor visitor = new MethodVisitor();
		AstBuilder._currentCompilationUnit.accept(visitor);
		for (MethodDeclaration method : visitor.getMethodDeclarations()) {
			this.method = method;
			IMethodBinding binding = method.resolveBinding();
			IMethod jdtMethod = null;
			if (binding != null) {
				jdtMethod = (IMethod) binding.getJavaElement();
				IMethodBinding overriddenMethod = Bindings.findOverriddenMethod(binding, true);
			}

			// if the method declaration is binary, it should not be parsed.
			// this is only required in case of experiment set up.
			// Eclipse does not allow to rename binary methods.
			// remove when done using experiment set up
			// if(AstBuilder._renamedElement==null && !isMethodBinary()){
			if (AstBuilder._renamedElement != null) {
				getRenamedMethodInfo();
				if (methodfound) {
					break;
				}
			}
		}
		parseMethodInvocations(visitor);
	}

	/**
	 * Get the properties required to parse the method. Properties- Method name
	 * ( need to check it against the renamed method to ensure the right method
	 * parsing) Index of the leading comment for the parsed method. Parent class
	 * name of the method begin parsed.
	 */
	private void getMethodProperties() {
		this.methodName = method.getName().toString();
		this.leadingCommentIndex = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(method);
		this.parentName = ParseAstNodes.getClassName(method.getParent());
	}

	@SuppressWarnings("restriction")
	private boolean isMethodOverridden() {
		IMethodBinding binding = method.resolveBinding();
		if (binding != null) {
			IMethodBinding overriddenMethod = Bindings.findOverriddenMethod(binding, true);
			if (overriddenMethod != null)
				return true;
		}
		return false;
	}

	@SuppressWarnings("restriction")
	private boolean isMethodBinary() {
		IMethodBinding binding = method.resolveBinding();
		if (binding != null) {
			IMethodBinding overriddenMethod = Bindings.findOverriddenMethod(binding, true);
			if (overriddenMethod != null) {
				ITypeBinding classBinding = overriddenMethod.getDeclaringClass();
				if (classBinding != null && !classBinding.isFromSource())
					return true;
			}
		}
		return false;
	}

	private void getRenamedMethodInfo() {
		getMethodProperties();
		if (methodName.equals(AstBuilder._renamedElement.getElementName()) && parentName != null
				&& !parameterName.isPresent()) {
			// if(parentName.equalsIgnoreCase(AstBuilder._identifier.get_className()))
			{
				setRenamedMethodProperties();
				methodfound = true;
			}
		} else if (parameterName != null && parameterName.isPresent()) {
			setRenamedParameterProperties();
			methodfound = true;
		}

	}

	private void setRenamedMethodProperties() {
		setMethodParamtersList(method, parameterName, AstBuilder._identifier);
		int classModifier = ParseAstNodes.getClassModifier(parentName, method.getParent());
		int mod = method.getModifiers();
		AstBuilder._identifier.set_type(Type.METHOD);
		AstBuilder._identifier.set_modifier(ParseAstNodes.mapModifier(mod));
		AstBuilder._identifier.set_parentModifier(ParseAstNodes.mapModifier(classModifier));
		AstBuilder._identifier.set_isOverridden(isMethodOverridden());
		AstBuilder._identifier.set_fullQualifiedName(method.getName().getFullyQualifiedName());
		// _identifier.set_cuName(_currentCompilationUnit.getJavaElement().getElementName());
		ParseAstNodes.traverseLeadingComments(leadingCommentIndex);
	}

	private void setRenamedParameterProperties() {
		String paramName = setMethodParamtersList(method, parameterName, AstBuilder._identifier);
		if (paramName != null) {
			AstBuilder._identifier.set_type(Type.LOCAL);
			ParseAstNodes.traverseLeadingComments(leadingCommentIndex);
		}
	}

	/**
	 * Initializes the parameters list for a method
	 * 
	 * @param method
	 *            the method for which we need the parameters list
	 * @param parameterName
	 *            the parameter name to match against
	 * @return returns the param name
	 */
	private String setMethodParamtersList(MethodDeclaration method, Optional<String> parameterName,
			Identifier identifier) {
		String paramName = null;
		List<String> idParamList = new ArrayList<String>();
		@SuppressWarnings("unchecked")
		List<SingleVariableDeclaration> params = method.parameters();
		if (params.size() > 0 && params.get(0) instanceof SingleVariableDeclaration) {
			for (SingleVariableDeclaration param : params) {
				idParamList.add(param.getName().toString());
				if (parameterName.isPresent() && param.getName().toString().equals(parameterName.get())) {
					paramName = param.getName().toString();
				}
			}

			if (idParamList.size() > 0 && identifier != null)
				identifier.set_paramterNames(idParamList);
		}
		return paramName;
	}

	/**
	 * Parses all the method invocation ASt nodes.
	 * 
	 * @param visitor
	 *            the visitor used to visit the AST nodes.
	 */
	private void parseMethodInvocations(MethodVisitor visitor) {
		for (MethodInvocation method : visitor.getMethodInvocations()) {
			String methodName = method.getName().toString();
			int index = AstBuilder._currentCompilationUnit.firstLeadingCommentIndex(method);
			if (AstBuilder._renamedElement != null
					&& methodName.equals(AstBuilder._renamedElement.getElementName())) {
				AstBuilder._identifier.set_type(Type.METHOD);
				ParseAstNodes.traverseLeadingComments(index);
			}
		}

	}
}
